#!/usr/bin/python
# coding=utf-8

"""    pwn2tabfile - converts PWN Oxford 2004/2006/2007 dictionary bases to StarDict format
    Copyright (C) 2008-2009  Michał "mziab" Ziąbkowski

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>."""

import sys, zlib, codecs, struct, re, os

try:
	import psyco
	psyco.full()
except ImportError:
	pass

# defines what gets changed to what
replacements = {"&amp;": "&",
		"&rsquo;" : "'",
		"&agrave;": u"\u00E0",
		"&ccedil;": u"\u00E7",
		"&egrave;": u"\u00E8",
		"&eacute;": u"\u00E9",
		"&ecirc;": u"\u00EA",
		"&reg;": u"\u00AE",
		"<IMG SRC=\"ipa503.JPG\">": u"\u02D0",
		"&IPA502;": u"\u02CC",
		"&inodot;": u"\u026A",
		"<IMG SRC=\"schwa.JPG\">": u"\u0259",
		"<IMG SRC=\"ipa306.JPG\">": u"\u0254",
		"<IMG SRC=\"ipa313.JPG\">": u"\u0252",
		"<IMG SRC=\"IPAa313.JPG\">": u"\u0252",
		"<IMG SRC=\"ipa314.JPG\">": u"\u028C",
		"<IMG SRC=\"ipa321.JPG\">": u"\u028A",
		"<IMG SRC=\"ipa305.JPG\">": u"\u0251",
		"<IMG SRC=\"ipa326.JPG\">": u"\u025C",
		"<IMG SRC=\"ipa182.JPG\">": u"\u0255",
		"&#952;": u"\u03B8",
		"&##952;": u"\u03B8",
		"&##8747;": u"\u0283",
		"<SUB><IMG SRC=\"ipa135.JPG\"></SUB>": u"\u0292",
		"&eng;": u"\u014B",
		"&##39;": "'",
		"<SUB><IMG SRC=\"rzym1.jpg\"></SUB>": "I",
		"<SUB><IMG SRC=\"rzym2.jpg\"></SUB>": "II",
		"<SUB><IMG SRC=\"rzym3.jpg\"></SUB>": "III",
		"<SUB><IMG SRC=\"rzym4.jpg\"></SUB>": "IV",
		"<SUB><IMG SRC=\"rzym5.jpg\"></SUB>": "V",
		"<SUB><IMG SRC=\"rzym6.jpg\"></SUB>": "VI",
		"<SUB><IMG SRC=\"rzym7.jpg\"></SUB>": "VII",
		"<SUB><IMG SRC=\"rzym8.jpg\"></SUB>": "VIII",
		"&square;": u"\u2026",
		"&quotlw;": "\"",
		"&quotup;": "\"",
		"<IMG SRC=\"idioms.JPG\">": "IDIOMS:",
		"&squareb;": u"\u2022",
		"&hfpause;": "-",
		"&tilde;": "~",
		"&aelig;": u"\u00E6",
		"&rarr;": u"\u2192",
		"&theta;": u"\u03B8",
		"&pause;": u"\u2014",
		"&lsquo;": u"\u2018",
		"&eth;": u"\u00F0",
		"&para;": u"\u00B6",
		"&deg;": u"\u00B0",
		"&iuml;": u"\u00EF",
		"&ouml;": u"\u00F6",
		"&ocirc;": u"\u00F4",
		"&acirc;": u"\u00E2",
		"&epsilon;": u"\u03B5",
		"&icirc;": u"\u00EE",
		"&uuml;": u"\u00FC",
		"&##163;": u"\u00A3",
		"&dots;": u"\u2026",
		"&rArr;": u"\u2192",
		"&IPA118;": u"\u0272",
		"&##949;": u"\u03B5",
		"<IMG SRC=\"ipa183.JPG\">": u"\u0291",
		"&IPA413;": u"\u031F",
		"&IPA424;": u"\u0303",
		"&IPA505;": u"\u0306",
		# added for PWN 2006/2007
		"&IPA135;": u"\u0292",
		"&IPA305;": u"\u0251",
		"&IPA306;": u"\u0254",
		"&IPA313;": u"\u0252",
		"&IPAa313;": u"\u0252",
		"&IPA314;": u"\u028C",
		"&IPA321;": u"\u028A",
		"&IPA326;": u"\u025C",
		"&IPA503;": u"\u02D0",
		"&IPA146;": "h",
		"&IPA170;": "w",
		"&IPA128;": "f",
		"&IPA325;": u"\u00E6",
		"&IPA301;": "i",
		"&IPA155;": "l",
		"&IPA319;": u"\u026A",
		"&IPA114;": "m",
		"&IPA134;": u"\u0283",
		"&IPA103;": "t",
		"&IPA140;": "x",
		"&IPA119;": u"\u014B",
		"&IPA131;": u"\u00F0",
		"&IPA130;": u"\u03B8",
		"&schwa.x;": u"\u0259",
		"&apos;": "'",
		"&pound;": u"\u00A3",
		"&epsi;": u"\u03B5",
		"&ldquor;": u"\u201C",
		"&rdquo;": u"\u201D",
		"&mdash;": u"\u2014",
		"&marker;": u"\u2022",
		"<SUB><IMG SRC=\"rzym9.jpg\"></SUB>": "IX",
		"<SUB><IMG SRC=\"rzym10.jpg\"></SUB>": "X",
		"<SUB><IMG SRC=\"rzym11.jpg\"></SUB>": "XI",
		"<SUB><IMG SRC=\"rzym12.jpg\"></SUB>": "XII",
		"<SUB><IMG SRC=\"rzym13.jpg\"></SUB>": "XIII",
		"<SUB><IMG SRC=\"rzym14.jpg\"></SUB>": "XIV",
		"<SUB><IMG SRC=\"rzym15.jpg\"></SUB>": "XV",
		"&IPA101;": "p",
		"&IPA102;": "b",
		"&IPA104;": "d",
		"&IPA109;": "k",
		"&IPA110;": "g",
		"&IPA116;": "n",
		"&IPA122;": "r",
		"&IPA129;": "v",
		"&IPA132;": "s",
		"&IPA133;": "z",
		"&IPA153;": "j",
		"&IPA182;": u"\u0255",
		"&IPA183;": u"\u0291",
		"&IPA302;": "e",
		"&IPA304;": "a",
		"&IPA307;": "o",
		"&IPA308;": "u",
		"&IPA309;": "y",
		"&IPA322;": u"\u0259",
		"&IPA426;": u"\u02E1",
		"&IPA491;": u"\u01EB",
		"&IPA501;": u"\u02C8",
		"&ldquo;": u"\u201C",
		"&ndash;": u"\u2013",
		"&hellip;": u"\u2026",
		"&asymp;": u"\u2248",
		"&aacute;": u"\u00E1",
		"&comma;": ",",
		"&squ;": u"\u2022",
		"&ncaron;": u"\u0148",
		"&arabicrsquo.x;": "'",
		"&atildedotbl.x;": u"\u00E3",
		# Added for PWN Oxford 2005
		"<LITERA SRC=\"ipa135.JPG\">": u"\u0292",
		"<LITERA SRC=\"ipa305.JPG\">": u"\u0251",
		"<LITERA SRC=\"ipa306.JPG\">": u"\u0254",
		"<LITERA SRC=\"ipa313.JPG\">": u"\u0252",
		"<LITERA SRC=\"ipa314.JPG\">": u"\u028C",
		"<LITERA SRC=\"ipa321.JPG\">": u"\u028A",
		"<LITERA SRC=\"ipa326.JPG\">": u"\u025C",
		"<LITERA SRC=\"ipa503.JPG\">": u"\u02D0",
		"<LITERA SRC=\"schwa.JPG\">": u"\u0259",
		"<LITERA SRC=\"ipa182.JPG\">": u"\u0255",
		"<LITERA SRC=\"ipa183.JPG\">": u"\u0291",
		"&Eacute;": u"\u00E9",
		"&auml;": u"\u00E4",
		"&rdquor;": u"\u201D",
		"&verbar;": "|",
		"&IPA405;": u"\u0324",
		"&idot;": u"\u0130",
		"&equals;": "=",
		"&vprime;": "'",
		"&lsqb;": u"\u005B",
		"&rsqb;": u"\u005D",
		# added for PWN Russian
		"&Acy;": u"\u0410",
		"&acy;": u"\u0430",
		"&Bcy;": u"\u0411",
		"&bcy;": u"\u0431",
		"&CHcy;": u"\u0427",
		"&chcy;": u"\u0447",
		"&Dcy;": u"\u0414",
		"&dcy;": u"\u0434",
		"&Ecy;": u"\u042d",
		"&ecy;": u"\u044d",
		"&Fcy;": u"\u0424",
		"&fcy;": u"\u0444",
		"&Gcy;": u"\u0413",
		"&gcy;": u"\u0433",
		"&HARDcy;": u"\u042A",
		"&hardcy;": u"\u044A",
		"&Icy;": u"\u0418",
		"&icy;": u"\u0438",
		"&IEcy;": u"\u0415",
		"&iecy;": u"\u0435",
		"&IOcy;": u"\u0401",
		"&iocy;": u"\u0451",
		"&Jcy;": u"\u0419",
		"&jcy;": u"\u0439",
		"&Kcy;": u"\u041a",
		"&kcy;": u"\u043a",
		"&KHcy;": u"\u0425",
		"&khcy;": u"\u0445",
		"&Lcy;": u"\u041b",
		"&lcy;": u"\u043b",
		"&Mcy;": u"\u041c",
		"&mcy;": u"\u043c",
		"&Ncy;": u"\u041d",
		"&ncy;": u"\u043d",
		"&numero;": u"\u2116",
		"&Ocy;": u"\u041e",
		"&ocy;": u"\u043e",
		"&Pcy;": u"\u041f",
		"&pcy;": u"\u043f",
		"&Rcy;": u"\u0420",
		"&rcy;": u"\u0440",
		"&Scy;": u"\u0421",
		"&scy;": u"\u0441",
		"&SHCHcy;": u"\u0429",
		"&shchcy;": u"\u0449",
		"&SHcy;": u"\u0428",
		"&shcy;": u"\u0448",
		"&SOFTcy;": u"\u042C",
		"&softcy;": u"\u044C",
		"&Tcy;": u"\u0422",
		"&tcy;": u"\u0442",
		"&TScy;": u"\u0426",
		"&tscy;": u"\u0446",
		"&Ucy;": u"\u0423",
		"&ucy;": u"\u0443",
		"&Vcy;": u"\u0412",
		"&vcy;": u"\u0432",
		"&YAcy;": u"\u042f",
		"&yacy;": u"\u044f",
		"&Ycy;": u"\u042b",
		"&ycy;": u"\u044b",
		"&YUcy;": u"\u042e",
		"&yucy;": u"\u044e",
		"&Zcy;": u"\u0417",
		"&zcy;": u"\u0437",
		"&ZHcy;": u"\u0416",
		"&zhcy;": u"\u0436",
		"&Delta;": u"\u0394",
		"&xutri;": u"\u25B3",
		"&diam;": u"\u22C4",
		"&Aacute;": u"\u00C1",
		"&Aring;": u"\u00C5",
		"&Ccaron;": u"\u010C",
		"&yacute;": u"\u00FD",
		"&rcaron;": u"\u0159",
		"&Icirc;": u"\u00CE",
		"&Omacr;": u"\u014C",
		"&Ouml;": u"\u00D6",
		"&Rcaron;": u"\u0158",
		"&ccaron;": u"\u010D",
		"&Scaron;": u"\u0160",
		"&scaron;": u"\u0161",
		"&Zcaron;": u"\u017D",
		"&zcaron;": u"\u017E",
		"&uacute;": u"\u00FA",
		"&oslash;": u"\u00F8",
		"&abreve;": u"\u0103",
		"&laquo;": u"\u00AB",
		"&raquo;": u"\u00BB",
		"&umacr;": u"\u016B",
		"&tcedil;": u"\u0162",
		"&iacute;": u"\u00ED",
		"&ecaron;": u"\u011B",
		"&euml;": u"\u00EB",
		"&ntilde;": u"\u00F1",
		"&scedil;": u"\u015F",
		"&omacr;": u"\u014D",
		"&igrave;": u"\u00EC",
		"&#337;": u"\u0150",
		"&amacr;": u"\u0101",
		"&ograve;": u"\u00F2",
		"&atilde;": u"\u00E3",
		"&uring;": u"\u016F",
		"&#8470;": u"\u2116",
		"&ucirc;": u"\u00FB",
		"&aring;": u"\u00E5",
		"&Ccedil;": u"\u00C7",
		"&amp;lt;": u"\u003C",
		"&amp;gt;": u"\u003E",
		"&amp;amp;": "&",
		"&lt;": u"\u003C"
		}

# attempts to emulate the sort routine from stardict
def stardict_strcmp(x, y):
	pos = 0
	string_x = x[0]
	string_y = y[0]
	while (pos<len(string_x) and pos<len(string_y)):
		c1 = ord(string_x[pos].lower())
		c2 = ord(string_y[pos].lower())

		if c1 != c2:
			return c1-c2

		pos += 1

	if pos<len(string_x):
		c1 = ord(string_x[pos])
	else:
		c1 = 0
	
	if pos<len(string_y):
		c2 = ord(string_y[pos])
	else:
		c2 = 0

	# special case: identical word, differing only in case
	# StarDict wants the ones with capital letters first
	if c1-c2==0:
		return ord(string_x[0])-ord(string_y[0])

	return c1-c2

# converts and cleans up definitions/words
def format(temp):
	# convert symbols to unicode
	for i in replacements.items():
		temp=temp.replace(i[0], i[1])
	
	# needs to be done here in this specific order
	temp=temp.replace("&ap;~", u"\u2248")
	temp=temp.replace("&ap;", u"\u2248")

	# some clean-up
	temp=re.sub("^<BIG>.*?</BIG> ?", "", temp)			# remove entry name, we already have it
	temp=re.sub("<ICON.*?>", "", temp)
	temp=re.sub("<P> ?</P>", "", temp) 				# remove redundant <P>s
	temp=temp.replace("<P>", "<br><br>") 				# change <P>s to linebreaks
	temp=temp.replace("</P><B>", "<br><br><B>") 			# add linebreak before some bold words
	temp=temp.replace("</P>", "")
	temp=temp.replace("<HANGINGPAR>", "")
	temp=re.sub("<TEXTSECTION.*?>","", temp)
	temp=re.sub("</?PL>", "", temp)
	temp=re.sub("</?GB>", "", temp)
	temp=re.sub("<SUP>(.*?)</SUP>", " \\1", temp)
	temp=re.sub("<HMS NR=\"(.*?)\">", " \\1", temp)
	# added for PWN 2006/2007
	temp=re.sub("</?STEM>", "", temp)
	temp=re.sub("</?PH>", "", temp)
	# added for PWN Russian
	temp=re.sub("</?RU>", "", temp)
	temp=re.sub("<IGNORE>.*?</IGNORE>", "", temp)
	temp=re.sub("<IGNORE>.*", "", temp)				# leftovers in rus-pol

	# fix Polish-English ugliness
	temp=temp.replace("\n", "<br>")
	temp=re.sub("</?A.*?>", "", temp)
	temp=re.sub("^<br><br> ?", "", temp)

	# remove leading and trailing whitespace
	temp=temp.strip()

	return temp

if len(sys.argv) != 3:
	print "Usage: %s pwn_file.win stardict_file.tab" % sys.argv[0]
	sys.exit(1)

try:
	dict = open(sys.argv[1], "rb")

except IOError:
	print "Couldn't open dictionary"
	sys.exit(1)

# global variables
wordcount = 0
index_base = 0
words_base = 0
encoding = ""
header_offset = 2

# check header
header = struct.unpack("<I",dict.read(4))[0]

# seek and read data
if header == 0x81125747:
	print "Detected Oxford PWN 2004/2005 format..."
	encoding = "cp1250"

	# checking for 2005 version
	if struct.unpack("<I", dict.read(4))[0] == 2:
		header_offset = 5

	dict.seek(0x18)
	wordcount, index_base, words_base = struct.unpack("<3I",dict.read(12))
elif header == 0x81135747:
	print "Detected PWN Russian format..."
	encoding = "iso-8859-2"
	dict.seek(0x18)
	wordcount, index_base, words_base = struct.unpack("<3I",dict.read(12))
elif header == 0x81145747:
	print "Detected Oxford PWN 2006/2007 format..."
	encoding = "iso-8859-2"
	dict.seek(0x68)
	wordcount, index_base, temp, words_base = struct.unpack("<4I",dict.read(16))
else:
	print "Incorrect file header"
	sys.exit(1)

offsets = []
words = []

# read alphabetical index
print "Reading alphabetical index..."

dict.seek(index_base)
for i in xrange(wordcount):
	addr = struct.unpack("<I",dict.read(4))[0] & 0x07ffffff
	offsets += [addr]

# read index table and definitions
print "Reading words and definitions..."

for i in xrange(wordcount):
	dict.seek(words_base + offsets[i])
	word_buffer=dict.read(30000)

	if ord(word_buffer[2+1]) not in [0x49, 0x38]: 		# 0x49, 0x38 = information entries
		word_buffer = word_buffer[2+4+6:]
		length = word_buffer.index("\x00")
		word = word_buffer[:length]

		word = word.decode(encoding)
		word = format(word)
		word = word.replace("|", "") 		# make Polish->English words nicer to look at
		word = re.sub(" \d+$", "" , word) 	# remove word-final numbers

		# read definition
		length = word_buffer.index("\x00")
		word_buffer = word_buffer[length+header_offset:]

		temp = ""

		if word_buffer=="":
			continue

		if ord(word_buffer[0]) < 20:
			# take pointer and decompress
			temp = zlib.decompress(word_buffer[ ord(word_buffer[0]) + 1:])
		else:
			# copy up to next null char
			pos = word_buffer.index("\x00")
			temp = word_buffer[:pos]
		
		cur = temp.decode(encoding)
		words += [(word, format(cur))]

dict.close()

# write data
print "Writing converted data..."

# sort words, since stardict requires it
words.sort(stardict_strcmp)

try:
	tab = codecs.open(sys.argv[2], "w", "UTF-8")

except IOError:
	print "Couldn't open output file"
	sys.exit(1)

for i in words:
	tab.write("%s\t%s\n" % (i[0], i[1]))

tab.close()

